/*  Monkey HTTP Daemon
 *  ------------------
 *  Copyright (C) 2001-2003, Eduardo Silva P.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/***********************************************/ 
/* Modulo dir_html.c escrito por Daniel R. Ome */
/***********************************************/

#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <time.h>
#include <unistd.h>

#include "monkey.h"

#define  DIRECTORIO		  "     -"

/* Longitud de la cadena */
#define  MAX_LEN_STR		 30

/* Longitud de la fecha y hora */
#define  MAX_TIME			 17

/* Longitud del tamao */
#define  MAX_SIZE			  6

/* Incremento */
#define  GROW				  100

#define  SPACE				 ' '
#define  ZERO				  '\0'

/* Estructura de lista de archivos y directorios */
struct f_list {
	char	  path[MAX_PATH+1];			/* Ruta de acceso               */
	char	  size[MAX_SIZE+1];			/* Tamao del archivo           */
	char	  ft_modif[MAX_TIME+1];	  /* Fecha y hora de modificacion */
};

/* Ordenar cadenas de caracteres por el metodo SHELL */
/* Cdigo tomado del libro de Kernighan y Ritchie    */
struct f_list *shell (struct f_list *b, int n)
{
	int				gap, i, j;
	struct f_list  temp;

	for (gap = n/2; gap > 0; gap /= 2)
		for (i = gap; i <= n; i ++)
			for (j = i-gap;
					j >= 0 && ((strcmp(b[j].path, b[j+gap].path))>0);
					j -= gap)
			{
				temp			  = b[j];
				b[j]			  = b[j+gap];
				b[j+gap]		 = temp;
			}

	return (struct f_list *)b;
}

/* Si encuentra un ' ' en la cadena lo reemplaza por su valor 
   en hexadecimal (%20). Autor: Eduardo Silva */
char *check_string(char *str)
{
	int pos, offset;
	char *buffer=0, *final_buffer=0;
	
	final_buffer=M_strdup(str);
	while((pos=str_search(final_buffer," ",1))!=-1){
		buffer=M_malloc(strlen(final_buffer)+3);
		memset(buffer, '\0', sizeof(buffer));

		strncpy(buffer, final_buffer, pos);
		strncpy(buffer+pos, "%20", 3);
		offset=pos+1;
		strncpy(buffer+pos+3, final_buffer+offset, strlen(final_buffer+offset));
		buffer[strlen(final_buffer)+2]='\0';
		
		final_buffer=M_strdup(buffer);
	}
	return (char *) final_buffer;
}

/* Recortar la cadena si excede el ancho de MAX_LEN_STR */
char *cut_string(char *str)
{
	int i, k, j, len;
	static char s[MAX_PATH];

	if ((len = strlen(str)) > MAX_LEN_STR) {
		k = MAX_LEN_STR/2 - 2;
		for (i=0; i<k; i++)
			s[i] = str[i];

		s[i++] = '.';
		s[i++] = '.';
		s[i++] = '.';
		j		= i;
		k		= len - k;

		for (i=k; str[i]; i++, j++)
			s[j] = str[i];

		s[j] = ZERO;
		return (char *) s;
	}

	return (char *) str;
}

/* Agregar un elemento del directorio */
struct f_list *add_element(struct f_list *object, char *string,
                              int *count, int *max, struct stat *buffer)
{
	off_t	tam = 0;
	time_t	*fecha;
	char	tipo;
	struct f_list  *bak;


	if ((tam = strlen(string)) >= MAX_PATH-5)
		return (struct f_list *) object;

	(*count) ++;                     /* Un elemento maxs ... */
                                    /* Si object->count = 0 es */
                                    /* el primer elemento      */

	if ((*count) != 0){              /* Aumentar el tamao del array */
		if ((*count) >= (*max)) {
			bak = (struct f_list *) realloc(object, (GROW+(*max)) * sizeof(struct f_list));
	
			if (bak == NULL)
				return NULL;

			(*max)+= GROW;
			object = bak;
		}
	}
	
	/* Guardar la fecha y hora del elemento */
	fecha = (time_t *) buffer->st_mtime;
	strftime(object[*count].ft_modif, 255, "%d-%b-%G %H:%M",
               (struct tm *)localtime((time_t *) &fecha));


	/* Es directorio o archivo ?? */
	if (S_ISDIR(buffer->st_mode)) {
		strncpy(object[*count].size, DIRECTORIO, MAX_SIZE);

		/* Colocar el slash de terminacion */
		string[tam++] = '/';
		string[tam]	= ZERO;
	} else {
		tam = buffer->st_size;

		if (tam < 9999) {
			tipo = 'b';
		} else
			if((tam /= 1024) < 9999) {
				tipo = 'K';
			} else
				if((tam /= 1024) < 9999) {
					tipo = 'M';
				} else
					tipo = 'G';

		sprintf(object[*count].size, "%5lu%c", tam, tipo);
	}

	/* Guardar el nombre del directorio o archivo para ingresar */
	strncpy(object[*count].path, string, MAX_PATH);
	object[*count].path[MAX_PATH] = ZERO;

	return (struct f_list *) object;
}

/* Envia informacion del directorio en formato HTML 
   Modificacion : 2002/10/22 
   -> Agregado Chunked Transfer Encoding support para HTTP/1.1
*/
int GetDir(char *dirname, int socket, char *path_request, 
	char *header_file, char *footer_file, struct log_info *s_log, int protocol)
{
	DIR *dir;
	struct dirent *ent;
	char *path, *real_header_file, *real_footer_file;
	struct stat *buffer;
	struct f_list *file_list;
	struct header_values *hd;
	
	int i,
 		count_file=-1,	  /* Cantidad de elementos        */
	     max_file=0,		  /* Cantidad tope usada por GROW */
		 transfer_type; /* Tipo de transferencia de datos */
		 
	if ((dir = opendir(dirname)) == NULL)
		return -1;

	if ((file_list = (struct f_list *) M_malloc(sizeof(struct f_list))) == NULL)
	{
		closedir(dir);
		return -1;
	}

	if ((path = (char *) calloc(sizeof(char), MAX_PATH)) == NULL)
	{
		free(file_list);
		closedir(dir);
		return -1;
	}

	if ((buffer = (struct stat *) M_malloc(sizeof(struct stat))) == NULL)
	{
		free(path);
		free(file_list);
		closedir(dir);
		return -1;
	}

	/* Leer los archivos y directorios */
	while ((ent = readdir(dir)) != NULL) {
		if (strcmp((char *) ent->d_name, "." )  == 0) continue;
		if (strcmp((char *) ent->d_name, ".." ) == 0) continue;

		snprintf(path, MAX_PATH, "%s/%s%c", dirname, ent->d_name, ZERO);

		if (stat(path, buffer) == -1) continue;

		if ((file_list = (struct f_list *) add_element(file_list, ent->d_name,
	                                                   &count_file, &max_file,
                                                      buffer)) == NULL) {
			free(path);
			free(buffer);
			closedir(dir);
			return -1;
		}
 	}

	/* Ordenar el arreglo de archivos y directorios */
	shell(file_list, count_file);

	hd = M_malloc(sizeof(struct header_values));
	hd->status = M_HTTP_OK;
	hd->content_length = 0;
	hd->content_type = m_build_buffer("text/html");
	hd->location = NULL;
	hd->cgi = SH_CGI;
	hd->pconnections_left = 0;
	
	if(protocol==HTTP_11){
		transfer_type=CHUNKED;
		M_METHOD_send_headers(socket, hd, s_log);
		fdprintf(socket, NO_CHUNKED, "Transfer-Encoding: Chunked\r\n\r\n");
	}
	else{
		transfer_type=NO_CHUNKED;
		M_METHOD_send_headers(socket, hd, s_log);
		fdprintf(socket, transfer_type, "\r\n");
	}
	free(hd);
	fdprintf(socket, transfer_type, 
		"<HTML>\n<HEAD><TITLE>Index of %s</TITLE></HEAD>\n <BODY> \
				 <H1>Index of %s</H1>", path_request, path_request);
				 
				 
	real_header_file = m_build_buffer("%s%s", dirname, header_file);
	if(AccessFile(real_header_file)==0){
		FILE *file;
		int bytes;
		char buffer[BUFFER_SOCKET];
		
		file = fopen(real_header_file, "r");
		memset(buffer, '\0', sizeof(buffer));
		while((bytes=fread(buffer,1, BUFFER_SOCKET, file)>0)){
			fdprintf(socket, transfer_type, "%s", buffer);
			memset(buffer, '\0', sizeof(buffer));
		}
		fclose(file);
	}
	
	fdprintf(socket, transfer_type,"<BR> <PRE>    Modified        Size     Name\n<HR> \
			 \n\t\t\t -   <A HREF=\"../\">Parent Directory</A>\n");
	

	for (i=0; i<=count_file; i++)
	{
		fdprintf(socket, transfer_type, " %s  %s   <A HREF=\"%s\">%s</A>\n",
                file_list[i].ft_modif, file_list[i].size,
                check_string(file_list[i].path)    , cut_string(file_list[i].path));
	}


	fdprintf(socket, transfer_type, "</PRE><HR>");
	
	real_footer_file = m_build_buffer("%s%s", dirname, footer_file);
	if(AccessFile(real_footer_file)==0){
		FILE *file;
		int bytes;
		char buffer[BUFFER_SOCKET];
		
		file = fopen(real_footer_file, "r");
		memset(buffer, '\0', sizeof(buffer));
		while((bytes=fread(buffer,1, BUFFER_SOCKET, file)>0)){
			fdprintf(socket, transfer_type, "%s", buffer);
			memset(buffer, '\0', sizeof(buffer));
		}
		fclose(file);
		fdprintf(socket, transfer_type, "<HR>");
		
	}
	
	fdprintf(socket, transfer_type, "<ADDRESS>%s</ADDRESS></BODY></HTML>\n", \
		config->server_signature);

	if(transfer_type==CHUNKED)
		fdprintf(socket, CHUNKED, "");
		
	free(file_list);
	free(path);
	free(buffer);
	free(real_header_file);
	free(real_footer_file);
	closedir(dir);

	return 0;
}
 

